# $Date: 1996/11/15 07:21:12 $ $Author: kg $ $Revision: 1.18.2.2 $ #

# kg, 09/12/93 #

#++
gcd -- compute the gcd of polynomials or expressions

gcd(poly,...)
gcd(expr,...)

poly - polynomials
expr - polynomial expressions

gcd computes the gcd of polynomials or expressions.
++#

gcd:= proc()
    local a, p, q, n, T, rand, mul, i, g, f, conv, normp, gcdp;
begin
    if args(0) = 0 then return(0) end_if;
    a:= { args() };
    
    # only integer or rational args? #
    T:= map(a, domtype);
    if (T minus {DOM_INT,DOM_RAT,DOM_FLOAT,DOM_COMPLEX}) = {} then
	if T = {DOM_INT} then return(igcd(op(a))) end_if;
	return(1)
    end_if;
    
    p:= op(a,1);
    n:= nops(a);
    # handle cases n=1 and n>2 #
    case n
    of 2 do break;
    of 1 do
	T:= op(T,1);
	if T::zero <> FAIL then return(gcd(p, T::zero)) end_if;
	if T = DOM_POLY then return(gcd(p, poly(0,op(p,2..3)))) end_if;
	return(gcd(p, 0));

    otherwise
        # get random generator for coeffs #
        if T = {DOM_POLY} then
            T:= op(p, 3);
            if T = hold(Expr) then
                rand:= random;
            elif domtype(T) = DOM_DOMAIN then
                rand:= T::random;
                if rand = FAIL then
                    # sort according to ascending degrees, then iterate gcd #
                    a:= sort([op(a)], fun(bool(degree(args(1)) < degree(args(2)))));
                    g:= a[1];
                    for i from 2 to n do
                        g:= gcd(g, a[i])
                    end_for;
                    return(g)
                end_if
            else # T = IntMod(q) #
		# create random number generator #
		q:= op(T,1);
		if not contains(gcdlib::gen_random, q) then
		    sysassign(gcdlib::gen_random[q], random(q))
		end_if;
		rand:= gcdlib::gen_random[q];
            end_if;
            mul:= multcoeffs;
        else
            rand:= random;
            mul:= _mult;
        end_if;

	# get gcd of several args by probabilitic approach, see Zippel, R.,
	  Effective Polynomial Comp., Kluwer 1993, p127 #
        f:= fun(divide(args(1), g, Exact) = FAIL);
        while n <> 1 do
	    if n = 2 then return(gcd(op(a))) end_if;
	    g:= n div 2;
	    p:= [ op(a, 1..g) ];
	    q:= [ op(a, (g+1)..n) ];
	    g:= gcd(_plus(p[1], mul(p[i], rand()) $ i=2..nops(p)),
		    _plus(q[1], mul(q[i], rand()) $ i=2..nops(q)));
	    a:= select(a, f) union {g};
	    n:= nops(a);
	end_while;
	return(op(a,1));
    end_case;

    # now n=2 #
    q:= op(a,2);
    # try to overload #
    if p::gcd <> FAIL then return(p::gcd(p,q)) end_if;
    if q::gcd <> FAIL then return(q::gcd(p,q)) end_if;

    # convert args to polys #
    if T = {DOM_POLY} then
	T:= op(p, 3);
	conv:= id;
	# test args #
	if testargs() then
	    if op(p, 2..3) <> op(q, 2..3) then
		error("polynomial types differ")
	    end_if;
	    if domtype(T) = DOM_DOMAIN then
		# need gcd method for coeffs #
		if T::gcd = FAIL then error("method 'gcd' missing") end_if;
		if T::divex = FAIL then error("method 'divex' missing") end_if
	    elif T = hold(Expr) then
		# can't handle real or complex coeffs #
		case gcdlib::coeff_type([p,q])
		of DOM_INT do
		of DOM_RAT do
		    break;
		otherwise
		    error("illegal coefficients")
		end_case
	    else # T = IntMod(p) #
		# polynomial with coeffs from Zp must have prime modulus #
		if not isprime(op(T,1)) then
		    error("modulus must be prime")
		end_if
	    end_if;
	end_if;
    else
	# convert expr to polys #
	T:= hold(Expr);
	conv:= expr;
	i:= indets(a, hold(PolyExpr));
	if (i = {}) or (i minus indets(i, hold(RatExpr)) <> {}) then
	    a:= op(a);
	    i:= map(a, normal);
	    if i = a then error("not a polynomial") end_if;
	    return(gcd(i))
	end_if;
	i:= [op(i)];
	p:= poly(p, i);
	q:= poly(q, i);
	if p = FAIL or q = FAIL then
	    a:= op(a);
	    i:= map(a, normal);
	    if i = a then error("not a polynomial") end_if;
	    return(gcd(i))
	end_if;
    end_if;
    
    if iszero(p) then return(conv(q)) end_if;
    if iszero(q) then return(conv(p)) end_if;

    f:= [];
    if T = hold(Expr) then
	# polynomial with expressions as coeffs #
	# change varibles which are expressions into indets,
	  otherwise scaling may cause problems later on #
	if select(map(op(p,2), domtype), _unequal, DOM_IDENT) <> [] then
	    f:= op(p,2);
	    g:= [ (if domtype(f[i]) = DOM_IDENT then f[i]
		   else genident() end_if) $ hold(i)=1..nops(f) ];
	    p:= poly(poly2list(p), g, T);
	    q:= poly(poly2list(q), g, T);
	end_if;
	T:= map({coeff(p), coeff(q)}, domtype);
	if T = {DOM_INT} then
	    gcdp:= gcdlib::int_gcd;
	    normp:= gcdlib::norm_intp;
	elif T minus {DOM_INT,DOM_RAT} = {} then
	    gcdp:= gcdlib::int_gcd;
	    normp:= fun(args(1));
	    conv:= conv @ fun(multcoeffs(args(1), 1/lcoeff(args(1))));
	    p:= multcoeffs(p, 1/icontent(p));
	    q:= multcoeffs(q, 1/icontent(q));
	else
	    # we try our best... #
	    if conv = expr then
		i:= [ mapcoeffs(p, normal), mapcoeffs(q, normal) ];
		if i = [p, q] then error("illegal coefficients") end_if;
		return(expr(gcd(op(i))))
	    end_if;
	    g:= gcd(expr(p), expr(q));
	    if g = FAIL then
		error("illegal coefficients")
	    end_if;
	    g:= poly(g, op(p, 2..3));
	    if f <> [] then g:= poly(poly2list(g), f, op(p,3)) end_if;
	    return(g);
	end_if
    elif domtype(T) = DOM_DOMAIN then
	gcdp:= gcdlib::dom_gcd;
	normp:= gcdlib::norm_domp;
    else # T = IntMod(p) #
	gcdp:= gcdlib::mod_gcd;
	normp:= gcdlib::norm_modp;
	conv:= fun(multcoeffs(args(1), 1/lcoeff(args(1))));
    end_if;
    
    g:= gcdlib::special_cases(p, q, gcdp, normp);
    if f <> [] then g:= poly(poly2list(g), f, op(p,3)) end_if;
    conv(g);
end_proc:

# end of file #
